Imports iNTERFACEWARE.Chameleon
Imports System.Reflection

' THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
' IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FIT-
' NESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
' iNTERFACEWARE BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
' IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CON-
' NECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

Public Class ClientModule
    Dim runningAsService As Boolean
    Dim logStream As System.IO.StreamWriter
    Dim client As LlpClient
    Dim state As ClientState
    Dim tickTimer As System.Windows.Forms.Timer

    Const APP_LOG_FILE As String = "hl7log.txt"
    Const APP_MAX_MESSAGE_DISPLAY_LENGTH As Integer = 2048
    Const APP_REMOTE_PORT As Integer = 5123
    Const APP_REMOTE_HOST As String = "localhost"
    Const APP_TICK_INTERVAL As Integer = 1000  ' interval in milliseconds of tick timer
    Const APP_ACK_TIMEOUT As Integer = 120 ' timeout for ACK in seconds

    Public Sub New(ByVal isService As Boolean)
        runningAsService = isService
        If (runningAsService) Then
            logStream = System.IO.File.AppendText(ExecutableDirectory() + APP_LOG_FILE)
        End If
    End Sub

    Public Function ExecutableDirectory() As String
        Return System.IO.Path.GetDirectoryName(System.Reflection.Assembly.GetExecutingAssembly().Location) + "\\"
    End Function

    Public Sub StartService()
        Try
            Log("Starting service.")
            client = New LlpClient
            AddHandler client.MessageReceived, AddressOf Client_MessageReceived
            AddHandler client.Closed, AddressOf Client_Closed
            AddHandler client.Connected, AddressOf Client_Connected
            AddHandler client.DataIgnored, AddressOf Client_DataIgnored
            AddHandler client.ErrorReceived, AddressOf Client_ErrorReceived

            tickTimer = New System.Windows.Forms.Timer
            tickTimer.Interval = APP_TICK_INTERVAL
            AddHandler tickTimer.Tick, AddressOf TickTimer_Tick
            tickTimer.Start()

            state = New ClientState

            ' Run the message loop.  This will run in an infinite loop until it receives a WM_QUIT message
            ' which is what it gets when client.StopMessageLoop() is invoked.
            ' If you wish to add functionality to the start up of the service be sure to add it before this
            ' line or do it on the timer tick events.
            LlpClient.RunMessageLoop()
            ' Cleaning up after the message loop has stopped
            client.Disconnect()
        Catch ex As Exception
            Log("Error starting service: " + ex.Message)
        End Try
    End Sub

    Public Sub StopService()
        Try
            Log("Stopped service")
            client.StopMessageLoop()
        Catch ex As Exception
            Log("Error stopping service:" + ex.Message)
        End Try
    End Sub

    Private Sub ConnectToRemoteHost()
        Log("Connecting to port" + Str(APP_REMOTE_PORT) + " on " + APP_REMOTE_HOST)
        client.Connect(APP_REMOTE_HOST, APP_REMOTE_PORT)
    End Sub

    Private Sub PollAndSendMessage()
        Log("Polling for message to send...")
        If state.FetchMessage Then
            Log("found id=" + state.MessageId)
            state.SetAckTimeOut(APP_ACK_TIMEOUT)
            ' Sending the message over the network can always result in an exception since the remote host
            ' can disconnect at any time.  The application behavior is safe since it should simply time out
            ' and then resend the message.
            TransmitMessageWithLog()
        Else
            Log("No message to send.")
        End If
    End Sub

    Private Sub TransmitMessageWithLog()
        Log("Sending:")
        Log(PrettyPrintMessage(state.Message))
        client.SendMessage(state.Message)
        Log("Sent!")
    End Sub

    Private Sub DoConnectedAction()
        Select Case state.CurrentState
            Case Hl7State.NEED_TO_POLL_APP
                PollAndSendMessage()
                Return
            Case Hl7State.WAITING_FOR_ACK
                If (state.IsTimedOut()) Then
                    Log("Timed out waiting for ACK for " + state.MessageId)
                    ' We need to resend the message - we reset the ACK timeout
                    ' TODO - we're just resending the message here - other possible implementations would be to disconnect
                    ' before trying to resend.

                    ' TODO - if we get many resends in a row it would probably make sense to keep a count and then
                    ' send an email or pager alert.
                    state.SetAckTimeOut(APP_ACK_TIMEOUT)
                    Log("Resending...")
                    ' and resend the message
                    TransmitMessageWithLog()
                End If
                Return
        End Select
    End Sub

    Private Sub TickTimer_Tick(ByVal Sender As System.Object, ByVal e As System.EventArgs)
        If (Not client.IsConnected) Then
            ConnectToRemoteHost()
        Else
            DoConnectedAction()
        End If
    End Sub

    Private Sub Client_MessageReceived(ByVal Sender As System.Object, ByVal e As System.EventArgs, ByVal message As String)
        Log("Received an ack")
        Log(PrettyPrintMessage(message))

        Select Case state.CurrentState
            Case Hl7State.WAITING_FOR_ACK
                ' TODO - we'll assume that the ACK is well formed and the message control ID matches the one we were expecting
                ' We could parse that out with Chameleon if we wanted to be very paranoid.  In a synchronous exchange it should
                ' be okay...
                ' The following command is *VERY IMPORTANT* it marks the message as being sent - the ClientState code should
                ' be written in a manner to never consider a message send until this application gives confirmation.
                Log(state.MessageId + " marked as sent.")
                state.ConfirmMessageSend()
                PollAndSendMessage()
                Return
            Case Hl7State.NEED_TO_POLL_APP
                ' TODO page alert
                Log("WARNING!  We were not expecting an ACK.")
                Return
        End Select
    End Sub

    Private Sub Client_Closed(ByVal Sender As System.Object, ByVal e As System.EventArgs)
        Log("Connection Closed")
    End Sub

    Private Sub Client_ErrorReceived(ByVal Sender As System.Object, ByVal e As System.EventArgs, ByVal description As String, ByVal code As Integer)
        Log("Error Received : " + PrettyPrintMessage(description))
    End Sub

    Private Sub Client_DataIgnored(ByVal Sender As System.Object, ByVal e As System.EventArgs, ByVal data As String)
        Log("Ignored Data received: ")
        Log(PrettyPrintMessage(data))
    End Sub

    Private Sub Client_Connected(ByVal Sender As System.Object, ByVal e As System.EventArgs)
        Log("Connection established with host")
        DoConnectedAction()
    End Sub

    Private Function PrettyPrintMessage(ByVal message As String) As String
        Return client.MakeMessagePrintable(message, Constants.vbCrLf, APP_MAX_MESSAGE_DISPLAY_LENGTH)
    End Function

    Public Sub Log(ByVal line As String)
        Dim logLine As String
        logLine = System.DateTime.Now.ToString() + " " + line
        If (runningAsService) Then
            logStream.WriteLine(logLine)
            logStream.Flush()
        Else
            System.Console.WriteLine(logLine)
        End If
    End Sub

End Class
